/*
    In this walk:
     1) We check the type of the identifier is match with the operation that associate with it.
        and check the experssion Type is match with the operation of that.
     2) Check Function call and return
     3) In Tuple Expression, all fields are defined in schema and in assignment of tuple
        the fields left hand side tuple is equal to right hand side.
 */
package wig.type;

import wig.*;
import wig.node.*;
import wig.analysis.*;
import wig.symbol.*;
import wig.compiler.*;
import java.util.*;

public class TypeWalk extends DepthFirstAdapter
{

    private SymbolTable GlobalTable;
    private Hashtable cstToScopeMap;
    private Hashtable typeTable;
    private Type functionType;


    /************************************************8
     typeTable:
       hashtable (keys=Node,out=Type).
       Basically, whenever depth first search
       goes out of a expression, I put the expression
       node into the hashtable along with what type it is.
       That way, if I have an expression within an expression
       I'm able to retrieve the inner expression.
       For example: (x<3)&&(x>-1)
         x will be marked as an int
         3 will be marked as an int
         -1 will be marked as an int
         x<3 will be marked as boolean
         x>-1 will be marked as boolean
         (x<3)&&(x>-1) will be marked as boolean

      functionType:
         whenever depth first search enters a function
         its functionType is set to the function Type
         for use when DFS is going out of a return
         statement.

    **********************************************/


    public TypeWalk(AnalysisDataManager manager)
    {
        GlobalTable = manager.getGlobalTable();
        cstToScopeMap = manager.getCstToScopeMap();
        typeTable = manager.getTypeTable();

    }


    /**********************************************
    * finds the closest scope to a node
    **********************************************/
    private SymbolTable getSymboltable(Node node)
    {
        while(!cstToScopeMap.containsKey(node))
        {
            node = node.parent();
            if(node==null)
                break;
        }
        if(node==null)
        {
            return GlobalTable;
        }
        return (SymbolTable) cstToScopeMap.get(node);
    }


    /*set function type when scoping into a function*/
    public void inAFunction(AFunction node)
    {
        TIdentifier id = node.getIdentifier();
        Symbol s = GlobalTable.seekSymbol(id.getText());
        if(!s.getKind().equals(Const.function))
        {
            PrintError.error("Type Error: Bad function name",node.getIdentifier().getLine());
        }
        functionType = s.getType();
    }


    /**** Fieldvalue *****/


    /*
     * make sure that all left side of = match
     * its corresponding schema.
     */
    public void outAFieldvalue(AFieldvalue node)
    {
        Node tmpnode;

        tmpnode = node.parent();
        while(!(tmpnode instanceof AAssignExp))
        {
            tmpnode = tmpnode.parent();
        }
        if(tmpnode==null)
        {
            PrintError.error("Type Error(10): tuple {...} is not on the right side of an = (assignment)",
                             node.getAssign().getLine());
        }

        /*
         * 1. Get the Assign node
         * 2. Get the lvalue of that
         * 3. Get the Type of Lvalue
         * 4. Get the schema symbol from the type of the lvalue
         * 5. Get the Symbol Table from the node of the schema
         */
        AAssignExp node_assign = (AAssignExp) tmpnode;
        PLvalue node_lvalue = node_assign.getLvalue();
        Type type_lvalue = (Type) typeTable.get(node_lvalue);
        Symbol symbol_schema = GlobalTable.seekSymbol(type_lvalue.getType());
        SymbolTable schematable = (SymbolTable) cstToScopeMap.get(symbol_schema.getNode());
        if(schematable==null)
        {
            PrintError.error("Type Error(11): tuple {...} is not proper",
                             node.getAssign().getLine());
            return;
        }
        Symbol s = schematable.seekSymbol(node.getIdentifier().getText());
        if(s==null)
        {
            PrintError.error("Type Error(12): tuple {..."+node.getIdentifier().getText()+
                             "=...} does not exitss in its corresponding schema",node.getAssign().getLine());
            return;
        }


        Type type_id = s.getType();
        Type type_exp = (Type) typeTable.get(node.getExp());

        //build up a type for the tuple{fieldvalus} operations
        tmpnode = node.parent();
        while(!(tmpnode instanceof ATupleBaseExp))
        {
            tmpnode = tmpnode.parent();
        }
        Type tupletype = (Type)typeTable.get(tmpnode);
        tupletype.addField(node.getIdentifier().getText(), type_exp);
        //set it back
        //End

        if(!type_id.equals(type_exp))
        {
            PrintError.error("Type Error(13): tuple {..."+node.getIdentifier().getText()+
                             "=...}  has a malformed type",node.getAssign().getLine());

        }
    }


    /**** Lvalue *****/

    /*
     * 1. make sure that the left id matches its corresponding
     *    delcared variable
     * 2. make sure that the right id matches its schemas type
     */
    public void outAQualifiedLvalue(AQualifiedLvalue node)
    {
        SymbolTable table = getSymboltable(node);
        String tuple_id = node.getLeft().getText();
        Symbol tuple_symbol = table.seekSymbol(tuple_id);
        Node n = node;
        while(tuple_symbol.getKind()==Const.hole)
        {
            /*we are inside an inputs, we want the second symbol table*/
            n = n.parent();
            table = getSymboltable(n);
            tuple_symbol = table.seekSymbol(tuple_id);
        }

        Type tuple_type = tuple_symbol.getType();

        // get the symbol of the schema
        Symbol schema_symbol = GlobalTable.seekSymbol(tuple_type.getType());
        if(schema_symbol==null)
        {
            PrintError.error("Type Error: schema type does not exist",
                             node.getLeft().getLine());
        }

        // get the schema symbol table by the node of the schema (SymbolTable)
        SymbolTable schema_table = (SymbolTable) cstToScopeMap.get(schema_symbol.getNode());
        if(schema_table==null)
        {
            PrintError.error("Type Error: schema type does not exist",
                             node.getLeft().getLine());
        }

        //check the field is defined and put the type into the typeTable
        String dotvar_id = node.getRight().getText();
        Symbol dotvar_symbol = schema_table.seekSymbol(dotvar_id);
        if(dotvar_symbol==null)
        {
            PrintError.error("Type Error: schema atrribute does not exist",
                             node.getLeft().getLine());
        }
        typeTable.put(node,dotvar_symbol.getType());

    }

    /*
     * make sure that the lvalue is the same type
     * as its delcaration
     */
    public void outASimpleLvalue(ASimpleLvalue node)
    {
        SymbolTable table = getSymboltable(node);
        String id = node.getIdentifier().getText();
        Node n = node;
        Symbol s = table.seekSymbol(id);
        while(s.getKind()==Const.hole)
        {
            /*we are inside an inputs, we want the second symbol table*/
            n = n.parent();
            table = getSymboltable(n);
            s = table.seekSymbol(id);
        }
        Type t = s.getType();
        typeTable.put(node,t);
    }



    /**** base_exp *****/

    public void outAParenBaseExp(AParenBaseExp node)
    {
        PExp exp = node.getExp();
        Type type = (Type) typeTable.get(node.getExp());
        typeTable.put(node,type);
    }

    //it should be inATupleBaseExp( to create the node first
    public void inATupleBaseExp(ATupleBaseExp node)
    {
        typeTable.put(node, new Type(""));
    }

    public void outAStringBaseExp(AStringBaseExp node)
    {
        typeTable.put(node,Const.tString);
    }

    public void outAFalseBaseExp(AFalseBaseExp node)
    {
        typeTable.put(node,Const.tBool);
    }

    public void outATrueBaseExp(ATrueBaseExp node)
    {
        typeTable.put(node,Const.tBool);
    }

    public void outAIntBaseExp(AIntBaseExp node)
    {
        typeTable.put(node,Const.tInt);
    }

    public void outACallBaseExp(ACallBaseExp node)
    {
        SymbolTable table = getSymboltable(node);
        Symbol s = table.seekSymbol(node.getIdentifier().getText());
        Type t = s.getType();
        typeTable.put(node,t);
    }


    public void outALvalueBaseExp(ALvalueBaseExp node)
    {
        PLvalue lvalue_node = node.getLvalue();
        Type t = (Type) typeTable.get(lvalue_node);
        typeTable.put(node,t);
    }

    /*********unary_exp***********************/

    /*
     * make sure that the exp is bool
     */
    public void outANotUnaryExp(ANotUnaryExp node)
    {
        PBaseExp base_exp = node.getBaseExp();
        Type type = (Type) typeTable.get(base_exp);
        if(!type.equals(Const.tBool))
        {
            PrintError.error("Type Error(9): ! operator applied non-bool value",
                             node.getNot().getLine());
        }
        else
        {
            typeTable.put(node,Const.tBool);
        }
    }


    /*
     * make sure that the exp is an int
     */
    public void outANegUnaryExp(ANegUnaryExp node)
    {
        PBaseExp base_exp = node.getBaseExp();
        Type type = (Type) typeTable.get(base_exp);
        if(!type.equals(Const.tInt))
        {
            PrintError.error("Type Error: - operator applied non-int value",
                             node.getMinus().getLine());
        }
        else
        {
            typeTable.put(node,Const.tInt);
        }
    }

    public void outADefaultUnaryExp(ADefaultUnaryExp node)
    {
        PBaseExp base_exp = node.getBaseExp();
        Type type = (Type) typeTable.get(base_exp);
        typeTable.put(node,type);
    }


    /*********tuple_exp***********************/
    // check the tuple operation of keep
    public void outAKeepTupleExp(AKeepTupleExp node)
    {
        PTupleExp tuple_exp = node.getTupleExp();
        Type oldtype = (Type) typeTable.get(tuple_exp);

        //protect the original tuple hashtable
        Type type=new Type(oldtype.getType());
        type.setFieldsTable(oldtype.cloneFieldsTable());

        if(type.equals(Const.tTuple))
        {
            String keepField = node.getIdentifier().getText();
            //update the hashtable inside the type object
            Type fieldType = type.getField(keepField);
            type.clearFields();

            if(fieldType != null)
                type.addField(keepField, fieldType);
            //else PrintError.error("Type Error: - field undefined",

            typeTable.put(node,type);
        }
        //else print out error
    }


    // check the tuple operation of remove
    public void outARemoveTupleExp(ARemoveTupleExp node)
    {
        PTupleExp tuple_exp = node.getTupleExp();
        Type oldtype = (Type) typeTable.get(tuple_exp);

        //protect the original tuple hashtable
        Type type=new Type(oldtype.getType());
        type.setFieldsTable(oldtype.cloneFieldsTable());

        if(type.equals(Const.tTuple))
        {
            String removeField = node.getIdentifier().getText();
            //update the hashtable inside the type object
            type.removeField(removeField);

            typeTable.put(node,type);
        }
        //else print out error
    }


    public void outAKeepManyTupleExp(AKeepManyTupleExp node)
    {
        PTupleExp tuple_exp = node.getTupleExp();
        Type oldtype = (Type) typeTable.get(tuple_exp);

        //protect the original tuple hashtable
        Type type=new Type(oldtype.getType());
        type.setFieldsTable(oldtype.cloneFieldsTable());

        Hashtable tempTable = new Hashtable();
        if(type.equals(Const.tTuple))
        {
            String keepField;
            Node temnode = node.getIdentifiers();
            while(temnode instanceof AManyIdentifiers)
            {
                keepField  = ((AManyIdentifiers)temnode).getIdentifier().getText(); //without s
                temnode = ((AManyIdentifiers)temnode).getIdentifiers(); // have s

                //put the keep fields into the temp hashtable first
                tempTable.put(keepField, type.getField(keepField));
            }
            keepField = ((AOneIdentifiers)temnode).getIdentifier().getText();
            tempTable.put(keepField, type.getField(keepField));

            type.setFieldsTable(tempTable);

            typeTable.put(node,type);
        }
    }


    public void outARemoveManyTupleExp(ARemoveManyTupleExp node)
    {
        PTupleExp tuple_exp = node.getTupleExp();
        Type oldtype = (Type) typeTable.get(tuple_exp);

        //protect the original tuple hashtable
        Type type=new Type(oldtype.getType());
        type.setFieldsTable(oldtype.cloneFieldsTable());

        if(type.equals(Const.tTuple))
        {
            String removeField;
            Node temnode = node.getIdentifiers();
            while(temnode instanceof AManyIdentifiers)
            {
                removeField  = ((AManyIdentifiers)temnode).getIdentifier().getText(); //without s
                temnode = ((AManyIdentifiers)temnode).getIdentifiers(); // have s

                //put the keep fields into the temp hashtable first
                type.removeField(removeField);
            }
            removeField = ((AOneIdentifiers)temnode).getIdentifier().getText();
            type.removeField(removeField);

            typeTable.put(node,type);
        }
    }

    public void outADefaultTupleExp(ADefaultTupleExp node)
    {
        PUnaryExp unary_exp = node.getUnaryExp();
        Type type = (Type) typeTable.get(unary_exp);
        typeTable.put(node,type);
    }


    /*********join_exp***********************/
    public void outAJoinJoinExp(AJoinJoinExp node)
    {
        PTupleExp left = node.getLeft();
        PJoinExp right = node.getRight();
        Type leftType = (Type) typeTable.get(left);
        Type rightType = (Type) typeTable.get(right);

        // add the symbol of the joined tuple into the rightSymTable too
        Hashtable joining = rightType.cloneFieldsTable();

        // when x<<y, replace the entry in x with y
        joining.putAll(leftType.getFieldsTable());
        Type type = new Type(leftType.getType(), joining);

        typeTable.put(node,type);
    }

    public void outADefaultJoinExp(ADefaultJoinExp node)
    {
        PTupleExp tuple_exp = node.getTupleExp();
        Type type = (Type) typeTable.get(tuple_exp);
        typeTable.put(node,type);
    }

    /*********mult_exp***********************/

    /*
     * 1. make sure that left and right are type int
     */
    public void outAMultMultExp(AMultMultExp node)
    {
        PMultExp left_exp = node.getLeft();
        Type left_type = (Type) typeTable.get(left_exp);
        if(!left_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: * operator applied to non-int value",
                             node.getMult().getLine());
        }
        PJoinExp right_exp = node.getRight();
        Type right_type = (Type) typeTable.get(right_exp);
        if(!right_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: * operator applied to non-int value",
                             node.getMult().getLine());
        }
        typeTable.put(node,Const.tInt);
    }


    public void outADivMultExp(ADivMultExp node)
    {
        PMultExp left_exp = node.getLeft();
        Type left_type = (Type) typeTable.get(left_exp);
        if(!left_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: / operator applied to non-int value",
                             node.getDiv().getLine());
        }
        PJoinExp right_exp = node.getRight();
        Type right_type = (Type) typeTable.get(right_exp);
        if(!right_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: / operator applied to non-int value",
                             node.getDiv().getLine());
        }
        typeTable.put(node,Const.tInt);
    }


    /*
     * 1. make sure that left and right are type int
     */
    public void outAModMultExp(AModMultExp node)
    {
        PMultExp left_exp = node.getLeft();
        Type left_type = (Type) typeTable.get(left_exp);
        if(!left_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: % operator applied to non-int value",
                             node.getMod().getLine());
        }
        PJoinExp right_exp = node.getRight();
        Type right_type = (Type) typeTable.get(right_exp);
        if(!right_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: % operator applied to non-int value",
                             node.getMod().getLine());
        }
        typeTable.put(node,Const.tInt);
    }



    public void outADefaultMultExp(ADefaultMultExp node)
    {
        PJoinExp join_exp = node.getJoinExp();
        Type type = (Type) typeTable.get(join_exp);
        typeTable.put(node,type);
    }

    /*********add_exp***********************/


    /* 1. Both sides of the plus must be the same type
     * 2. Both sides must either be an int or string
     */
    public void outAPlusAddExp(APlusAddExp node)
    {
        PAddExp left_exp = node.getLeft();
        Type left_type = (Type) typeTable.get(left_exp);
        if(!left_type.equals(Const.tInt)&&!left_type.equals(Const.tString))
        {
            PrintError.error("Type Error: + operator applied to non-valid types (int,string)",
                             node.getPlus().getLine());
        }
        PMultExp right_exp = node.getRight();
        Type right_type = (Type) typeTable.get(right_exp);
        if(!right_type.equals(Const.tInt)&&!right_type.equals(Const.tString))
        {
            PrintError.error("Type Error: + operator applied to non-valid types (int,string)",
                             node.getPlus().getLine());
        }
        if(!right_type.equals(left_type))
        {
            PrintError.error("Type Error: + operator applied to different type",
                             node.getPlus().getLine());
        }

        typeTable.put(node,left_type);
    }

    /*
     * 1. make sure that left and right are type int
     */
    public void outAMinusAddExp(AMinusAddExp node)
    {
        PAddExp left_exp = node.getLeft();
        Type left_type = (Type) typeTable.get(left_exp);
        if(!left_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: - operator applied to non-int value",
                             node.getMinus().getLine());
        }
        PMultExp right_exp = node.getRight();
        Type right_type = (Type) typeTable.get(right_exp);
        if(!right_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: - operator applied to non-int value",
                             node.getMinus().getLine());
        }
        typeTable.put(node,Const.tInt);
    }


    public void outADefaultAddExp(ADefaultAddExp node)
    {
        PMultExp mult_exp = node.getMultExp();
        Type type = (Type) typeTable.get(mult_exp);
        typeTable.put(node,type);
    }


    /*********cmp_exp***********************/


    /*
     * 1. make sure that left and right are the same type
     * 2. and set the output to bool
     */
    public void outAEqCmpExp(AEqCmpExp node)
    {
        PAddExp left_exp = node.getLeft();
        Type left_type = (Type) typeTable.get(left_exp);
        if(!left_type.equals(Const.tInt)&&!left_type.equals(Const.tString)
                &&!left_type.equals(Const.tBool)
                &&!left_type.equals(Const.tVoid))
        {
            PrintError.error("Type Error: == applied to non-valid types",
                             node.getEq().getLine());
        }
        PAddExp right_exp = node.getRight();
        Type right_type = (Type) typeTable.get(right_exp);

        if(!right_type.equals(Const.tInt)&&!right_type.equals(Const.tString)
                &&!right_type.equals(Const.tBool)
                &&!right_type.equals(Const.tVoid))
        {
            PrintError.error("Type Error: == applied to non-valid types",
                             node.getEq().getLine());
        }
        if(!right_type.equals(left_type))
        {
            PrintError.error("Type Error: == left side and right side are not the same type",
                             node.getEq().getLine());
        }
        typeTable.put(node,Const.tBool);
    }

    /*
     * 1. make sure that left and right are the same type
     * 2. and set the output to bool
     */
    public void outANeqCmpExp(ANeqCmpExp node)
    {
        PAddExp left_exp = node.getLeft();
        Type left_type = (Type) typeTable.get(left_exp);
        if(!left_type.equals(Const.tInt)&&!left_type.equals(Const.tString)
                &&!left_type.equals(Const.tBool)
                &&!left_type.equals(Const.tVoid))
        {
            PrintError.error("Type Error(6): != applied to non-valid types",
                             node.getNeq().getLine());
        }
        PAddExp right_exp = node.getRight();
        Type right_type = (Type) typeTable.get(right_exp);
        if(!right_type.equals(Const.tInt)&&!right_type.equals(Const.tString)
                &&!right_type.equals(Const.tBool)
                &&!right_type.equals(Const.tVoid))
        {
            PrintError.error("Type Error(7): != applied to non-valid types",
                             node.getNeq().getLine());
        }
        if(!right_type.equals(left_type))
        {
            PrintError.error("Type Error(8): != left side and right side are not the same type",
                             node.getNeq().getLine());
        }
        typeTable.put(node,Const.tBool);
    }


    /*
     * 1. make sure that left and right are both int
     * 2. and set the output to bool
     */
    public void outAGtCmpExp(AGtCmpExp node)
    {
        PAddExp left_exp = node.getLeft();
        Type left_type = (Type) typeTable.get(left_exp);
        if(!left_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: > operator applied to non-int value",
                             node.getGt().getLine());
        }
        PAddExp right_exp = node.getRight();
        Type right_type = (Type) typeTable.get(right_exp);
        if(!right_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: > operator applied to non-int value",
                             node.getGt().getLine());
        }
        typeTable.put(node,Const.tBool);
    }


    /*
     * 1. make sure that left and right are both int
     * 2. and set the output to bool
     */
    public void outALtCmpExp(ALtCmpExp node)
    {
        PAddExp left_exp = node.getLeft();
        Type left_type = (Type) typeTable.get(left_exp);
        if(!left_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: < operator applied to non-int value",
                             node.getLt().getLine());
        }
        PAddExp right_exp = node.getRight();
        Type right_type = (Type) typeTable.get(right_exp);
        if(!right_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: < operator applied to non-int value",
                             node.getLt().getLine());
        }
        typeTable.put(node,Const.tBool);
    }

    /*
     * 1. make sure that left and right are both int
     * 2. and set the output to bool
     */
    public void outALteqCmpExp(ALteqCmpExp node)
    {
        PAddExp left_exp = node.getLeft();
        Type left_type = (Type) typeTable.get(left_exp);
        if(!left_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: >= operator applied to non-int value",
                             node.getLteq().getLine());
        }
        PAddExp right_exp = node.getRight();
        Type right_type = (Type) typeTable.get(right_exp);
        if(!right_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: >= operator applied to non-int value",
                             node.getLteq().getLine());
        }
        typeTable.put(node,Const.tBool);
    }

    /*
     * 1. make sure that left and right are both int
     * 2. and set the output to bool
     */
    public void outAGteqCmpExp(AGteqCmpExp node)
    {
        PAddExp left_exp = node.getLeft();
        Type left_type = (Type) typeTable.get(left_exp);
        if(!left_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: <= operator applied to non-int value",
                             node.getGteq().getLine());
        }
        PAddExp right_exp = node.getRight();
        Type right_type = (Type) typeTable.get(right_exp);
        if(!right_type.equals(Const.tInt))
        {
            PrintError.error("Type Error: <= operator applied to non-int value",
                             node.getGteq().getLine());
        }
        typeTable.put(node,Const.tBool);
    }

    /*
     * 1. make sure that left and right are both int
     * 2. and set the output to bool
     */
    public void outADefaultCmpExp(ADefaultCmpExp node)
    {
        PAddExp add_exp = node.getAddExp();
        Type type = (Type) typeTable.get(add_exp);
        typeTable.put(node,type);
    }

    /*********and_exp***********************/

    /*
     * 1. make sure that left and right are both bool
     * 2. and set the output to bool
     */
    public void outAAndAndExp(AAndAndExp node)
    {
        PAndExp left_exp = node.getLeft();
        Type left_type = (Type) typeTable.get(left_exp);
        if(!left_type.equals(Const.tBool))
        {
            PrintError.error("Type Error: and operator applied to non-bool value",
                             node.getAnd().getLine());
        }
        PCmpExp right_exp = node.getRight();
        Type right_type = (Type) typeTable.get(right_exp);
        if(!right_type.equals(Const.tBool))
        {
            PrintError.error("Type Error: and operator applied to non-bool value",
                             node.getAnd().getLine());
        }
        typeTable.put(node,Const.tBool);
    }


    public void outADefaultAndExp(ADefaultAndExp node)
    {
        PCmpExp cmp_exp = node.getCmpExp();
        Type type = (Type) typeTable.get(cmp_exp);
        typeTable.put(node,type);
    }


    /*********or_exp***********************/

    /*
     * 1. make sure that left and right are both bool
     * 2. and set the output to bool
     */
    public void outAOrOrExp(AOrOrExp node)
    {
        POrExp left_exp = node.getLeft();
        Type left_type = (Type) typeTable.get(left_exp);
        if(!left_type.equals(Const.tBool))
        {
            PrintError.error("Type Error: or operator applied to non-bool value",
                             node.getOr().getLine());
        }
        PAndExp right_exp = node.getRight();
        Type right_type = (Type) typeTable.get(right_exp);
        if(!right_type.equals(Const.tBool))
        {
            PrintError.error("Type Error: or operator applied to non-bool value",
                             node.getOr().getLine());
        }
        typeTable.put(node,Const.tBool);
    }


    public void outADefaultOrExp(ADefaultOrExp node)
    {
        PAndExp and_exp = node.getAndExp();
        Type type = (Type) typeTable.get(and_exp);
        typeTable.put(node,type);
    }


    /*********exp***********************/

    /*
     *1. The case when the right side is a tuple
      *  is handled in base_exp {tuple}
     *2. do the left and right side have the same type
     */
    public void outAAssignExp(AAssignExp node)
    {
        PLvalue node_lvalue = node.getLvalue();
        Type type_lvalue = (Type) typeTable.get(node_lvalue);
        POrExp node_orexp = node.getOrExp();
        Type type_orexp = (Type) typeTable.get(node_orexp);
        /* Tuple can also be compared in here see Type.java*/
        if(!type_lvalue.equals(type_orexp))
        {
            PrintError.error("Type Error: = (assignment) left side type doesn't match right side type",
                             node.getAssign().getLine());
        }

    }



    public void outADefaultExp(ADefaultExp node)
    {
        POrExp or_exp = node.getOrExp();
        Type type = (Type) typeTable.get(or_exp);
        typeTable.put(node,type);
    }


    /********** input ***********/

    /*********** stm and stm_no_short_if *****************/

    /****make sure return; has void function type ****/
    public void outAReturnStm(AReturnStm node)
    {
        Type type = functionType;
        if(!type.equals(Const.tVoid))
        {
            PrintError.error("Type Error: void function has non-void return",
                             node.getReturn().getLine());
        }

    }

    /****make sure return; has void function type ****/
    public void outAReturnStmNoShortIf(AReturnStmNoShortIf node)
    {
        Type type = functionType;
        if(!type.equals(Const.tVoid))
        {
            PrintError.error("Type Error: void function has non-void return",
                             node.getReturn().getLine());
        }

    }

    /****make sure return type matches function ****/
    public void outARetexpStm(ARetexpStm node)
    {
        Type type = functionType;
        Type type_exp = (Type) typeTable.get(node.getExp());
        if(type.equals(Const.tVoid))
        {
            PrintError.error("Type Error: non-void function has void return",
                             node.getReturn().getLine());
        }
        if(!type.equals(type_exp))
        {
            PrintError.error("Type Error: return type doesn't match function type",
                             node.getReturn().getLine());
        }
    }

    /****make sure return type matches function ****/
    public void outARetexpStmNoShortIf(ARetexpStmNoShortIf node)
    {
        Type type = functionType;
        Type type_exp = (Type) typeTable.get(node.getExp());
        if(type.equals(Const.tVoid))
        {
            PrintError.error("Type Error: non-void function has void return",
                             node.getReturn().getLine());
        }
        if(!type.equals(type_exp))
        {
            PrintError.error("Type Error: return type doesn't match function type",
                             node.getReturn().getLine());
        }
    }


    /****make sure if statement exp is boolean ****/
    public void outAIfStm(AIfStm node)
    {
        PExp exp = node.getExp();
        Type type_exp = (Type) typeTable.get(exp);
        if(!type_exp.equals(Const.tBool.getType()))
        {
            PrintError.error("Type Error: if has non-bool type",
                             node.getIf().getLine());
        }

    }


    /****make sure if statement exp is boolean ****/
    public void outAIfelseStm(AIfStm node)
    {
        PExp exp = node.getExp();
        Type type_exp = (Type) typeTable.get(exp);
        if(!type_exp.equals(Const.tBool.getType()))
        {
            PrintError.error("Type Error: if has non-bool type",
                             node.getIf().getLine());
        }

    }

    /****make sure while statement exp is boolean ****/
    public void outAWhileStmNoShortIf(AWhileStmNoShortIf node)
    {
        PExp exp = node.getExp();
        Type type_exp = (Type) typeTable.get(exp);
        if(!type_exp.equals(Const.tBool))
        {
            PrintError.error("Type Error(4): while has non-bool type",
                             node.getWhile().getLine());
        }
    }

    /****make sure while statement exp is boolean ****/
    public void outAWhileStm(AWhileStm node)
    {
        PExp exp = node.getExp();
        Type type_exp = (Type) typeTable.get(exp);
        if(!type_exp.equals(Const.tBool))
        {
            PrintError.error("Type Error(5): while has non-bool type",
                             node.getWhile().getLine());
        }

    }

    /***make sure exit's html identifier exists ***/
    public void outAExitStm(AExitStm node)
    {
        PDocument pdoc = node.getDocument();
        String id = "";
        if(pdoc instanceof AIdDocument)
        {
            AIdDocument doc = (AIdDocument) pdoc;
            id = doc.getIdentifier().getText();
        }
        else if(pdoc instanceof APlugDocument)
        {
            APlugDocument doc = (APlugDocument) pdoc;
            id = doc.getIdentifier().getText();
        }
        
        // check the html const is defined
        if(!GlobalTable.defsSymbol(id))
        {
            PrintError.error("Type Error: "+id+"is not defined as an html variable",
                             node.getExit().getLine());
            return;
        }
        Symbol s = GlobalTable.seekSymbol(id);
        if(!Const.html.equals(s.getKind()))
        {
            PrintError.error("Type Error: "+id+"is not defined as an html variable",
                             node.getExit().getLine());

        }
    }

    /***make sure exit's html identifier exists ***/
    public void outAExitStmNoShortIf(AExitStmNoShortIf node)
    {
        PDocument pdoc = node.getDocument();
        String id = "";
        if(pdoc instanceof AIdDocument)
        {
            AIdDocument doc = (AIdDocument) pdoc;
            id = doc.getIdentifier().getText();
        }
        else if(pdoc instanceof APlugDocument)
        {
            APlugDocument doc = (APlugDocument) pdoc;
            id = doc.getIdentifier().getText();
        }
        if(!GlobalTable.defsSymbol(id))
        {
            PrintError.error("Type Error: "+id+"is not defined as an html variable",
                             node.getExit().getLine());
            return;
        }
        Symbol s = GlobalTable.seekSymbol(id);
        if(!Const.html.equals(s.getKind()))
        {
            PrintError.error("Type Error: "+id+"is not defined as an html variable",
                             node.getExit().getLine());
        }
    }


    /***make sure show's html identifier exists ***/
    public void outAShowStm(AShowStm node)
    {
        PDocument pdoc = node.getDocument();
        String id = "";
        if(pdoc instanceof AIdDocument)
        {
            AIdDocument doc = (AIdDocument) pdoc;
            id = doc.getIdentifier().getText();
        }
        else if(pdoc instanceof APlugDocument)
        {
            APlugDocument doc = (APlugDocument) pdoc;
            id = doc.getIdentifier().getText();
        }
        if(!GlobalTable.defsSymbol(id))
        {
            PrintError.error("Type Error: "+id+"is not defined as an html variable",
                             node.getShow().getLine());
            return;
        }
        Symbol s = GlobalTable.seekSymbol(id);
        if(!Const.html.equals(s.getKind()))
        {
            PrintError.error("Type Error: "+id+"is not defined as an html variable",
                             node.getShow().getLine());

        }
    }

    /***make sure show's html identifier exists ***/
    public void outAShowStmNoShortIf(AShowStmNoShortIf node)
    {
        PDocument pdoc = node.getDocument();
        String id = "";

        if(pdoc instanceof AIdDocument)
        {
            AIdDocument doc = (AIdDocument) pdoc;
            id = doc.getIdentifier().getText();
        }
        else if(pdoc instanceof APlugDocument)
        {
            APlugDocument doc = (APlugDocument) pdoc;
            id = doc.getIdentifier().getText();
        }

        if(!GlobalTable.defsSymbol(id))
        {
            PrintError.error("Type Error: "+id+"is not defined as an html variable",
                             node.getShow().getLine());
            return;
        }
        Symbol s = GlobalTable.seekSymbol(id);
        if(!Const.html.equals(s.getKind()))
        {
            PrintError.error("Type Error: "+id+"is not defined as an html variable",
                             node.getShow().getLine());

        }
    }
}
